# Generated by Django 3.0.6 on 2020-07-29 20:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_appointment", "0021_auto_20191024_1000"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="appointment",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("timepoint", "visit_code_sequence"),
            },
        ),
    ]
