# Generated by Django 2.0 on 2018-01-02 21:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("edc_appointment", "0012_auto_20171128_2153")]

    operations = [migrations.DeleteModel(name="Holiday")]
